/* Symbols needed for libmate-panel-applet-private-mini, which is used by the test
 * program */
#include <glib.h>
gboolean mate_panel_applet_frame_dbus_load (const gchar *iid, gpointer frame_act);
gboolean mate_panel_applet_frame_dbus_load (const gchar *iid, gpointer frame_act) { return FALSE; }
