/*
 * libwnck based pager Apple.
 * (C) 2001 Alexander Larsson
 *
 * Authors: Alexander Larsson
 *
 */

#ifndef __WORKSPACE_SWITCHER_APPLET_H__
#define __WORKSPACE_SWITCHER_APPLET_H__

#include <glib.h>
#include <mate-panel-applet.h>

#ifdef __cplusplus
extern "C" {
#endif

gboolean workspace_switcher_applet_fill(MatePanelApplet* applet);

#ifdef __cplusplus
}
#endif

#endif
